!+
! Module: rgssources_mod
!
! Authors:	Peter Videler - ESTEC/SOC
!		Ian Stewart - SSC/LUX
!          
! Description: Construct table with source positions to be used in RGS pipeline.
! 
! Comments:
!
!	TODO:
!	- test old list for empty labels?
!	- even more rigorous testing of old list??
!	- quit under any circumstances if old rgs list doesn't pass checks.
!	- write EXPOSURE kwd - see JO mail of 25 April.
!	- squawk if newEpicAttrNum > 99 (because then the keyword will be > 8 char long??? extend keyword rather!!??)
!	- can we get rid of the necessity to have --changeprime=yes as well as --withusersource=yes and userasprime=yes?
!	- take out dataMode=spectroscopy from oal set state (or elsewhere) and see what happens. (See CE mail of 27 July.)
!	- what about updates of post-5.1 files which nevertheless have not the full set of header keywords, or wrong disp/xdsp, etc?
!	- Change column names in kwds (eg FLAG* written to the table by dataFilter) in cases where the old name no longer exists?
!	- in checks of old file, make sure prime src index is sensible. User might have deleted rows or altered the number.
!	- Allow the task to extract the instrument and expsoure from rgsproc-format file names.
!	- When primestyle=label, it can't find the default PROPOSAL. String dump to check for trailing nulls etc.
!	- Use the second, non-modifying form of doFilter call when filtering epic list?
!	- EXP_ID keyword will have the wrong value (3-digit exp str instead of 13-digit obs+exp str) for a lot of pre-sas_5.3 rgs source lists. Change the update module so it patches these instances up.
!	- Rationalise --epicset (and srclist??) parameters - should they both be of type t instead of e?
!	- ras, decs and pas - parameters should be of type u (angle) but need to wait until the param interface is updated so that default settings can be properly identified.
!	- See mail JDR 12 Mar 2002. Split xdsp 'slit' width selection into 2: one of +-3 arcmin (or, even better, specified by a user parameter) to select sources; the other smaller, intended for pcms use, to make sure that prime source always lies right on the chip.
!
! $Id: rgssources_mod.f90,v 1.5 2000/08/11 13:08:45 pvideler Exp $
!- 
module rgssources_mod

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.
  use dal
  use cal
  use errorHandling
  use oal		! includes 'use caloaldefs'
  use constants
  use pointing_mod	! -> SkyPositionType
  use dummycal_mod
  use dataFilter
  use rgssources_init	! -> sourceType
use temp_ssclib
  use confusion_mod
  use coordinate_mod

use rgssources_shared	! If rgssources_update ever becomes a separate task,
use rgssources_update	! all the stuff in shared should go back -> main.

  implicit none 

  character(255), parameter :: epicBandsConfFileName = "epicenergybands.conf"
  character(255) :: fovFilterStr
  type(SpacecraftAttitudeType) :: scAttitude, rgsAttitude
  ! type defined in caloalutils/src/caloaldefs.f90
  integer(int32) ::&
    paramInstrumentId
  character(4) ::&
    paramExposureIdStr
  real(single), parameter ::&
    detUnitArcsec = 0.05 ! The unit of the DETX,Y system.
!*** should be in constants module
  real(double) ::&
    xPixelSizeWorld,&
    yPixelSizeWorld,&
    pixelAreaArcsec
  logical :: haveScAttitude

! Command-line parameters:
  logical ::&
    withEpicSet,&
    addUserSrc,&
    changePrimeSrc,&
    changeAttitude,&
    clobberOnLabelClash,&
    setUserSrcToPrime,&
    enableSpatialFilter,&
    writeObsAttributes,&
    writeExpAttributes,&
    processUserSrc,&
    bkgExcludeUserSrc,&
    doEpicConfusion
  character(len=255) ::&
    oldListSetName,&
    newListSetName,&
    epicSrcSetName,&
    atthkSetName,&
    primeSrcExpression,&
paramInstExpIdStr
  character(len=9) ::&
    userSrcStyle,&
    primeSrcStyle,&
    attitudeStyle
  character(len=20) ::&
    fileMode
  character(len=labelsize) ::&
    userSrcLabel,&
    primeSrcLabel
  character(len=labelsize-5) ::&
    epicLabelPrefix
  real(single) ::&
    userSrcRate,&
    userSrcRaDeg,&
    userSrcDecDeg,&
    userSrcXdspArcmin,&
    userSrcDispArcmin,&
    userAttitudeRaDeg,&
    userAttitudeDecDeg,&
    userAttitudeAposDeg,&
    bandMinEeV,&
    bandMaxEeV,&
    fracBackground
  integer(kind=int16) ::&
    paramIndexOfPrime

!integer :: devId
!real :: polyx(4), polyy(4)

  private 

  public rgssources

contains 

  !---------------------------------------------------------------------
  subroutine rgssources
  !---------------------------------------------------------------------

!use caloaldefs
use odfdesignators

    implicit none 

    integer(int16) :: maxNumSrcs, actualNumSrcs, newPrimeSrcIndex
    type(sourceType), dimension(:), allocatable :: newListSrcData
    type(DataSetT) :: set
    type(TableT) :: tab
    integer(int32) :: instrumentId, epicInstrumentId, newEpicAttrNum, numRows
    character(len=5) :: instrumentName
    type(ColumnT) ::&
      confnCol
    real(single), dimension(:), pointer ::&
      confnColPtr
    character(255) :: instExpIdStr
    real(double) :: epicTimeStamp

    type(InstrumentExposureIdentifierT) :: paramInstExpIdStructure
    ! type defined in caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start:

    call readParameters
    paramInstExpIdStructure&
    = instrumentExposureIdentifier(trim(paramInstExpIdStr))
    paramInstrumentId = instrument(paramInstExpIdStructure)
    paramExposureIdStr = exposureId(paramInstExpIdStructure) 

    if (trim(fileMode) .eq. 'modify') then
      call convertOldFiles(&
        oldListSetName,&
        writeObsAttributes,&
        writeExpAttributes,&
        paramInstrumentId,&
        paramInstExpIdStr& 
      )
    end if

    call getInstrument(instrumentId, instrumentName)
    call getInstExpId(instrumentId, instExpIdStr)
    call CAL_setState(instrument=instrumentId)
    call getPointing(instrumentId, instExpIdStr) ! sets global variables
    ! scAttitude and rgsAttitude.

    call calcMaxNumSrcs(maxNumSrcs)
!    if (enableSpatialFilter) call constructSelectionExpr
if (enableSpatialFilter) fovFilterStr = '(DELTA_XDSP in [-2.9:2.9])'

!.......................................................................
! Construct the source list in memory:

    allocate(newListSrcData(maxNumSrcs))

    call fillOutputSrcList(newListSrcData, actualNumSrcs, newEpicAttrNum)

!.......................................................................
! Now write (most of) the source list to file:

    call writeToDataSet(newListSrcData(1:actualNumSrcs), instrumentName&
    , newEpicAttrNum, instExpIdStr, set, tab)

!.......................................................................
! Last bits and pieces - FOV flagging, prime source, rga-plane confusion,
! and checks:

    ! Flag on whether passes FOV filter or not, then copy the column back
    ! into newListSrcData.
    !
    if (enableSpatialFilter) then
      ! NOTE! Epic sources which fall outside the FOV have already been
      ! screened out at this stage. Hence the only sources for which the
      ! flag value should be != 0 are proposal, pointing, previous rgs
      ! and user sources. If the pointing is not within the FOV there is
      ! something badly wrong!
      !
      call initFilter('isNull(DELTA_XDSP) || !'//fovFilterStr)
      numRows = doSelectModify(tab, 'FLAG', flagBitPos=OUTOFFOV_BIT)
      call reportOofovFlags(tab, newListSrcData(1:actualNumSrcs))
!***** warnings on newListSrcData%flag are at present done within this sub - this should really be made more global.
    end if

    if (withEpicSet .and. doEpicConfusion) then
      call getEpicAttData(epicInstrumentId, epicTimeStamp)
      call calcEpicPlaneConfusion(epicInstrumentId, epicTimeStamp&
      , newListSrcData(1:actualNumSrcs))

      ! NOTE!! The cal state is reset to the relevant epic camera inside
      ! this sub.

      call flagConfusedSources(tab, newListSrcData(1:actualNumSrcs))
    end if

    newPrimeSrcIndex = calcPrimeIndex(tab, newListSrcData(1:actualNumSrcs))

    if (trim(fileMode) .eq. 'create' .or. changePrimeSrc) then
      call setAttribute(tab, 'PRIMESRC', newPrimeSrcIndex)
    end if

    call calcRgsPlaneConfusion(newListSrcData(1:actualNumSrcs)&
    , newPrimeSrcIndex)

!***** check now that no columns have nulls in them except EPIC_FILE.

    ! Write rgs-plane confusion column to file.
    confnCol = column(tab, 'CONFUSION')
    confnColPtr => real32Data(confnCol)
    confnColPtr = newListSrcData(1:actualNumSrcs)%confusion
    call release(confnCol)

!    call deleteColumn(tab, 'FLAG') ! because there isn't one yet in the disk
!    ! file format.

    call message(AppMsg, Verbose, 'Closing source list.')

    call release(set)

  end subroutine rgssources

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    use param  

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

    paramInstExpIdStr = stringParameter('instexpid')
    writeObsAttributes = booleanParameter('writeobskwds')
    writeExpAttributes = booleanParameter('writeexpkwds')

    fileMode = stringParameter('filemode')
    if (trim(fileMode) .eq. 'modify') then
      oldListSetName = stringParameter('srclist')
      changePrimeSrc = booleanParameter('changeprime')
      changeAttitude = booleanParameter('changeattitude')
    else
      newListSetName = stringParameter('srclist')
      changePrimeSrc = .true.
    end if
    clobberOnLabelClash = booleanParameter('clobberonlabel')
    withEpicSet = booleanParameter('withepicset')
    if (withEpicSet) then
      epicSrcSetName = stringParameter('epicset')
      epicLabelPrefix = stringParameter('epiclabelprefix')
      doEpicConfusion = booleanParameter('doconfusion')
!!!      fracBackground = realParameter('bkgfraction')
fracBackground = 10.0
    end if
    addUserSrc = booleanParameter('addusersource')
    if (addUserSrc) then
      userSrcLabel = ucase(stringParameter('label'))
      userSrcRate = realParameter('rate')
      setUserSrcToPrime = booleanParameter('userasprime')
      userSrcStyle = stringParameter('positionstyle')
      if (userSrcStyle .eq. 'radec') then
        userSrcRaDeg = realParameter('ra')
        userSrcDecDeg = realParameter('dec')
      else ! userSrcStyle .eq. 'wrtatt'
        userSrcDispArcmin = realParameter('deltadisp')
        userSrcXdspArcmin = realParameter('deltaxdsp')
      end if
      processUserSrc    = booleanParameter('process')
      bkgExcludeUserSrc = booleanParameter('bkgexclude')
    else
      setUserSrcToPrime = .false.
    end if
    primeSrcStyle = stringParameter('primestyle')
    select case(primeSrcStyle)
    case('label')
      primeSrcLabel = ucase(stringParameter('primelabel'))
    case('index')
      paramIndexOfPrime = intParameter('primeindex')
    case('expression')
      primeSrcExpression = stringParameter('primeexpression')
    end select
    attitudeStyle = stringParameter('attstyle')
    if (attitudeStyle .eq. 'user') then
      userAttitudeRaDeg = realParameter('attra')
      userAttitudeDecDeg = realParameter('attdec')
      userAttitudeAposDeg = realParameter('attapos')
    else if (attitudeStyle .eq. 'mean' .or. attitudeStyle .eq. 'median') then
      atthkSetName = stringParameter('atthkset')
    end if

    enableSpatialFilter = booleanParameter('enablefilter')

!    call openParameters(trim(epicBandsConfFileName))
!if (parameterCount('bandloev') .lt. 2 .or. parameterCount('bandhiev') .lt. 2)&
!call error('tooFewEpicBands', 'Too few band limits found in the file '&
!//trim(epicBandsConfFileName)//'.')
!    bandMinEeV = realParameter('bandloev', epicBandNum - 1)
!    bandMaxEeV = realParameter('bandhiev', epicBandNum - 1)
!    call closeParameters()
!*******
bandMinEeV = 500.0
bandMaxEeV = 2000.0

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine getInstrument(instrumentId, instrumentName)
  !---------------------------------------------------------------------

    implicit none 

    integer(int32), intent(out) :: instrumentId
    character(*),   intent(out) :: instrumentName

    type(DataSetT) :: oldListSet

    !-------------------------------------------------------------------
    ! Start:

    if (trim(fileMode) .eq. 'modify') then
      oldListSet = dataSet(oldListSetName, READ)
      instrumentName = stringAttribute(oldListSet, 'INSTRUME')
      call release(oldListSet)
      instrumentId = instId(instrumentName)

      if (instrumentId .ne. paramInstrumentId) then
        call warning('instIdClash'&
        , 'The instrument recorded in the source list header keyword &
        &INSTRUME is not the same as that derived from the value of &
        &the --instexpid parameter.')
      end if

    else
      instrumentId = paramInstrumentId
      instrumentName = instId(instrumentId)
    end if

  end subroutine getInstrument

  !---------------------------------------------------------------------
  subroutine getInstExpId(instrumentId, instExpIdStr)
  !---------------------------------------------------------------------

    implicit none 

    integer(int32), intent(in) :: instrumentId
    character(*), intent(out) :: instExpIdStr

    type(DataSetT) :: oldListSet
    character(4) :: exposureIdStr

    !-------------------------------------------------------------------
    ! Start:

    if (trim(fileMode) .eq. 'modify') then
      oldListSet = dataSet(oldListSetName, READ)
      exposureIdStr = stringAttribute(oldListSet, 'EXPIDSTR')
      call release(oldListSet)

      if (exposureIdStr .ne. paramExposureIdStr) then
        call warning('expIdStrClash'&
        , 'The 4-character exposure identifier recorded in the source &
        &list header keyword EXPIDSTR is not the same as that derived &
        &from the value of the --instexpid parameter.')
      end if

      if (instrumentId .eq. RGS1) then
        instExpIdStr = 'R1'//exposureIdStr
      else if (instrumentId .eq. RGS2) then
        instExpIdStr = 'R2'//exposureIdStr
      else
call error('bug0', 'Contact the code developer.')
      end if

    else
      instExpIdStr = paramInstExpIdStr
    end if

  end subroutine getInstExpId

  !---------------------------------------------------------------------
  subroutine getPointing(instrumentId, instExpIdStr)
  !---------------------------------------------------------------------

    use pointing_mod

    implicit none 

    integer(int32), intent(in) ::&
      instrumentId
    character(*), intent(in) :: instExpIdStr

    real(double) ::&
      startTime
    type(DataSetT) :: atthkSet, oldListSet
    type(ExposureInfoType) ::&
      exposureInfo ! typedef from oal

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Obtaining pointing/attitude.')

    haveScAttitude = .false. ! default

    if (trim(fileMode) .eq. 'modify' .and. .not. changeAttitude) then ! get the
    ! attitude from the old rgs source list:
      oldListSet = dataSet(oldListSetName, READ)
      rgsAttitude%ra   = real64Attribute(oldListSet, 'RA_PNT')  * piOn180
      rgsAttitude%dec  = real64Attribute(oldListSet, 'DEC_PNT') * piOn180
      rgsAttitude%apos = real64Attribute(oldListSet, 'PA_PNT')  * piOn180
      call release(oldListSet)

  return
    end if

    if (attitudeStyle .eq. 'user') then
    ! NOTE!! The following checks are designed mostly to detect when the user has allowed the attra, attdec and attapos parameters to retain their default values. The user should not do this if attstyle has been set to 'user'. The checks depend to some degree on the values of the relevant parameter defaults, hence IF THESE ARE CHANGED THIS ALGORITHM MAY ALSO NEED TO BE CHANGED.
      if (userAttitudeRaDeg .lt. 0.0 .or. userAttitudeRaDeg .gt. 360.0) then
call error('illegalAttRa', 'The value of --attra, the user-specified &
&attitude ra, is either < 0  or > 360 degrees. Have you forgotten to set it?')
      end if
      if (userAttitudeDecDeg .lt. -90.0 .or. userAttitudeDecDeg .gt. 90.0) then
call error('illegalAttDec', 'The value of --attdec, the user-specified &
&attitude declination, is either < -90 or > 90 degrees. Have you forgotten &
&to set it?')
      end if
      if (userAttitudeAposDeg .lt. 0.0 .or. userAttitudeAposDeg .gt. 360.0) then
call error('illegalAttApos', 'The value of --attapos, the user-specified &
&attitude position angle, is either < 0  or > 360 degrees. Have you forgotten &
&to set it?')
      end if

      rgsAttitude%ra   = userAttitudeRaDeg * piOn180
      rgsAttitude%dec  = userAttitudeDecDeg * piOn180
      rgsAttitude%apos = userAttitudeAposDeg * piOn180

  return
    end if

!    call OAL_setState(instrument=instrumentId, exposureId=exposureIdStr)
!!*** change this.
!! IMS 10 Jan 2002 - why???
call OAL_setState(instrument=instrumentId, instrExposureId=instExpIdStr)

    call RGSLIB_refreshCache() ! in rgslib/dummycal_mod
    call OAL_exposureInfo(exposureInfo)
    startTime = ToTimeTag(exposureInfo%actualStartTime) ! routine from
    ! caloalutils/caloaldefs

    select case (attitudeStyle) 
    case ('start') 
      if (.not. OAL_getAttitude(startTime, scAttitude)) then
call error('noAttitude', 'No attitude information is available for the &
&start time of the exposure.')
      endif

    case ('mean') ! see task atthkgen
      atthkSet = dataSet(atthkSetName , READ)
      scAttitude%ra   = real64Attribute(atthkSet,"AAHFRA")  * piOn180
      scAttitude%dec  = real64Attribute(atthkSet,"AAHFDEC") * piOn180
      scAttitude%apos = real64Attribute(atthkSet,"AAHFPA")  * piOn180
      call release(atthkSet)

    case ('median') ! see task atthkgen
      atthkSet = dataSet(atthkSetName , READ)
      scAttitude%ra   = real64Attribute(atthkSet,"MAHFRA")  * piOn180
      scAttitude%dec  = real64Attribute(atthkSet,"MAHFDEC") * piOn180
      scAttitude%apos = real64Attribute(atthkSet,"MAHFPA")  * piOn180
      call release(atthkSet)

    case default ! note, can't be 'user' because we have dealt with that above.
call error('badAttitudeStyle', 'The value '//trim(attitudeStyle)//' of &
&parameter --attstyle is not recognised.')
    end select

    haveScAttitude = .true.

    ! Correct reference frame for boresight:
    rgsAttitude = instrumentPointing(scAttitude, time=startTime)
    ! routine from rgslib/pointing_mod. Needs CAL_setState hence SAS_CCF.

  end subroutine getPointing

  !---------------------------------------------------------------------
  subroutine calcMaxNumSrcs(maxNumSrcs)
  !---------------------------------------------------------------------

    implicit none 

    integer(int16), intent(out) :: maxNumSrcs

    integer(int16) :: numOldListSrcs, numNewEpicSrcs
    type(DataSetT) :: oldListSet, epicSrcSet
    type(TableT) :: oldListTab, epicSrcTab

    !-------------------------------------------------------------------
    ! Start:

    numOldListSrcs = 0
    if (trim(fileMode) .eq. 'modify') then
      ! Open oldListSet and get number of rows:
      oldListSet = dataSet(oldListSetName, READ)
      oldListTab = table(oldListSet, 'SRCLIST', READ)
      numOldListSrcs = numberOfRows(oldListTab)
      call release(oldListSet)
    end if

    numNewEpicSrcs = 0
    if (withEpicSet) then
      ! Open epicSrcSet and get number of rows:
      epicSrcSet = dataSet(epicSrcSetName, READ)
      epicSrcTab = table(epicSrcSet, 'SRCLIST', READ)
      numNewEpicSrcs = numberOfRows(epicSrcTab)
      call release(epicSrcSet)
    end if

    maxNumSrcs = numOldListSrcs + numNewEpicSrcs + 2 ! 2 for the proposal and
    ! onaxis sources.

    if (addUserSrc) maxNumSrcs = maxNumSrcs + 1

if (maxNumSrcs .le. 0) call error('bug1', 'Contact the code developer.')

  end subroutine calcMaxNumSrcs

  !---------------------------------------------------------------------
  subroutine fillOutputSrcList(newListSrcData, lastNewListSrc, newEpicAttrNum)
  !---------------------------------------------------------------------

    implicit none 

    type(sourceType), dimension(:), intent(inout) ::&
      newListSrcData
    integer(int16), intent(out) ::&
      lastNewListSrc
    integer(int32), intent(out) ::&
      newEpicAttrNum

    integer(int16) ::&
      numSrcs
    type(DataSetT) :: oldListSet, tempEpicSrcSet
    type(TableT) :: oldListTab, tempEpicSrcTab
    type(sourceType), dimension(:), allocatable ::&
      oldListSrcData, newEpicSrcData
    type(sourceType), dimension(1) ::&
      userSrcData
    integer(int16) ::&
      src,&
      src2,&
      numOldListSrcs,&
      maxNumNewEpicSrcs,&
      lastEpicSrc,&
      firstSrcToCheck,&
      previousLastNewListSrc,&
      onAxisNumber,&
      maxPreviousIndex

type(ColumnT) ::&
dispCol,&
xdspCol
real(single), dimension(:), pointer ::&
dispColPtr,&
xdspColPtr
real(double), dimension(:), pointer ::&
raPtr,&
decPtr
type(sourceType) ::&
dummySrcData

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Constructing the source list in memory.')

    numSrcs = size(newListSrcData)

    call initSrcDataStructure(newListSrcData)

    select case(trim(fileMode))
    case('create')
      call addProposalSrc(newListSrcData(1))
      call addOnaxisSrc(  newListSrcData(2))
      lastNewListSrc = 2

      newEpicAttrNum = 1

    case('modify')
!.......................................................................
! Check and load the old source list into memory:

      oldListSet = dataSet(oldListSetName, READ)
      oldListTab = table(oldListSet, 'SRCLIST', READ)
      numOldListSrcs = numberOfRows(oldListTab)
      allocate(oldListSrcData(numOldListSrcs))
      call initSrcDataStructure(oldListSrcData)
      call readTable(oldListTab, oldListSrcData)
      call release(oldListSet)

      call checkOldList(oldListSrcData, onAxisNumber)
!********

      ! Load the old list sources into the new array:
      firstSrcToCheck = 1
      lastNewListSrc = 0
      call loadAndTestForLabelClashes(oldListSrcData&
      , firstSrcToCheck, newListSrcData, lastNewListSrc)

      if (changeAttitude) call addOnaxisSrc(newListSrcData(onAxisNumber))

      newEpicAttrNum = 0
      do src = 1, lastNewListSrc
        if (newListSrcData(src)%epicFileAttrNum .ne. myInt32null) then
          newEpicAttrNum = max(newEpicAttrNum&
          , newListSrcData(src)%epicFileAttrNum)
        end if
      end do
      newEpicAttrNum = newEpicAttrNum + 1

    case default
call error('badFileMode', 'The value '//trim(fileMode)//' of parameter &
&--filemode is not recognised.')
    end select ! fileMode.

! At present, we have a newList sources array containing at least the PROPOSAL and ONAXIS at indices 1 and 2, and the remainder of the old list if there is one. So far all the labels of the new list entries are guaranteed to be unique. Next job is to load the new epic sources (if withEpicSet) into a separate array.

    if (withEpicSet) then
!.......................................................................
! Load epic sources into memory:

      tempEpicSrcSet = clone(epicSrcSetName, 'tempEpicSet.ds', TEMP)
!tempEpicSrcSet = clone(epicSrcSetName, 'tempEpicSet.ds', MODIFY)
!Can use this alternative if it is desired to keep the filtered epic set.

    xPixelSizeWorld = real64Attribute(tempEpicSrcSet, 'TCDELT1')
    yPixelSizeWorld = real64Attribute(tempEpicSrcSet, 'TCDELT2')
    pixelAreaArcsec = abs(xPixelSizeWorld * yPixelSizeWorld&
    * 3600.0 * 3600.0)

      tempEpicSrcTab = table(tempEpicSrcSet, 'SRCLIST', MODIFY)
dispCol = addColumn(tempEpicSrcTab, 'DELTA_DISP',   REAL32)
xdspCol = addColumn(tempEpicSrcTab, 'DELTA_XDSP',   REAL32)
dispColPtr => real32Data(dispCol)
xdspColPtr => real32Data(xdspCol)
raPtr => real64Data(column(tempEpicSrcTab, 'RA'))
decPtr => real64Data(column(tempEpicSrcTab, 'DEC'))

do src = 1, numberOfRows(tempEpicSrcTab)
  dummySrcData%raDeg  = raPtr( src - 1)
  dummySrcData%decDeg = decPtr(src - 1)
  call calculateFOVcoords(dummySrcData, rgsAttitude)
  dispColPtr(src - 1) = dummySrcData%deltaDispArcmin
  xdspColPtr(src - 1) = dummySrcData%deltaXdspArcmin
end do

call release(tempEpicSrcTab)
tempEpicSrcTab = table(tempEpicSrcSet, 'SRCLIST', MODIFY)

      maxNumNewEpicSrcs = spatialFilter(tempEpicSrcTab)
      allocate(newEpicSrcData(maxNumNewEpicSrcs))

      call readEpicData(tempEpicSrcTab, newEpicSrcData, lastEpicSrc)
!call deleteColumn(tempEpicSrcTab, 'DELTA_DISP')
!call deleteColumn(tempEpicSrcTab, 'DELTA_XDSP')
      call release(tempEpicSrcSet)

      newEpicSrcData%epicFileAttrNum = newEpicAttrNum

      do src = 1, lastEpicSrc - 1
        do src2 = src + 1, lastEpicSrc
          if (newEpicSrcData(src)%label .eq. newEpicSrcData(src2)%label) then
call error('nonUniqueEpicId', 'Two sources in the supplied EPIC list &
&have been found to have duplicate values of the *_ID_SRC column, resulting &
&in duplication of the label '//trim(newEpicSrcData(src)%label)//'.')
          end if
        end do
      end do
!***** why isn't this test absorbed into loadAndTestForLabelClashes?

      firstSrcToCheck = 1
      previousLastNewListSrc = lastNewListSrc
      call loadAndTestForLabelClashes(newEpicSrcData(1:lastEpicSrc)&
      , firstSrcToCheck, newListSrcData, lastNewListSrc)

      if (lastNewListSrc - previousLastNewListSrc .le. 0) then
        call warning('noValidEpicSources', 'None of the epic sources have &
        &met all the conditions for inclusion.')
      end if
    end if ! withEpicSet

    if (addUserSrc) then
!.......................................................................
! Load user source into memory:
      call getUserSrc(userSrcData(1))

      firstSrcToCheck = 1
      call loadAndTestForLabelClashes(userSrcData&
      , firstSrcToCheck, newListSrcData, lastNewListSrc)
    end if

!.......................................................................
! Now the new or output source list is complete, in that all the sources
! are present. All label values are unique. We have still to fill in or
! correct a few columns though.

    ! ra/dec or fov:
    ! Note: all sources (except all previous rgs sources in 'modify' mode,
    ! and the ONAXIS source in 'create' mode) should have either ra/dec or
    ! xdsp/disp/phi/r set to (my)nulls at this point. After the following loop
    ! none of these 6 columns should have any nulls in them.
    do src = 1, lastNewListSrc
      if (newListSrcData(src)%fixedOnSky) then ! ra and dec should be left
      ! untouched. Do we recalculate the others?
        if (changeAttitude&
        .or. newListSrcData(src)%deltaDispArcmin .eq. myReal32null&
        .or. newListSrcData(src)%deltaXdspArcmin .eq. myReal32null) then ! yes.
          call calculateFOVcoords(newListSrcData(src), rgsAttitude)
        end if

      else ! not 'fixed on sky', the disp/xdisp and fov values should not
      ! be touched. Do we recalculate ra and dec?
        if (changeAttitude&
        .or. newListSrcData(src)%raDeg .eq. myReal32null&
        .or. newListSrcData(src)%decDeg .eq. myReal32null) then ! yes.
          call invertFOVcoords(newListSrcData(src), rgsAttitude)
        end if
      end if
    end do

    ! Calculate INDEX value. First find the maximum non-null existing value:
    maxPreviousIndex = 0
    do src = 1, lastNewListSrc
      if (newListSrcData(src)%index .ne. myInt16null) then
        maxPreviousIndex = max(maxPreviousIndex, newListSrcData(src)%index)
      end if
    end do
    ! Now fill in all the null values:
    do src = 1, lastNewListSrc
      if (newListSrcData(src)%index .eq. myInt16null) then
        maxPreviousIndex = maxPreviousIndex + 1
        newListSrcData(src)%index = maxPreviousIndex
      end if
    end do

! Note: can't do confusion until after prime source selected.

  end subroutine fillOutputSrcList

  !---------------------------------------------------------------------
  subroutine checkOldList(oldListSrcData, onAxisNumber)
  !---------------------------------------------------------------------

    implicit none 

    type(sourceType), dimension(:), intent(in) ::&
      oldListSrcData
    integer(int16), intent(out) :: onAxisNumber
    logical ::&
      oldListProposalExists,&
      oldListOnaxisExists
    integer(int16) :: src, proposalNumber, numOldListSrcs

    !-------------------------------------------------------------------
    ! Start:

    numOldListSrcs = size(oldListSrcData)

    ! Do some checks of the PROPOSAL source in the old list: 
    oldListProposalExists = .false.
    do src = 1, numOldListSrcs
      if (oldListSrcData(src)%label .eq. proposalLabel) then
        if (oldListProposalExists) then ! there is more than one PROPOSAL
        ! in oldListSrcData. Call error, regardless of the state of
        ! clobberOnLabelClash.
call error('tooManyProposals', 'There is more than one proposal source &
&in the old RGS source list.')

        else
          oldListProposalExists = .true.
          proposalNumber = src
        end if
      end if
    end do
if (.not. oldListProposalExists) call error('noProposal', 'There is &
&no proposal source in the old RGS source list.')

    ! Do some checks of the ONAXIS source in the old list: 
    oldListOnaxisExists = .false.
    do src = 1, numOldListSrcs
      if (oldListSrcData(src)%label .eq. onAxisLabel) then
        if (oldListOnaxisExists) then ! there is more than one ONAXIS
        ! in oldListSrcData. Call error, regardless of the state of
        ! clobberOnLabelClash.
call error('tooManyAttSrc', 'There is more than one pointing source &
&in the old RGS source list.')

        else
          oldListOnaxisExists = .true.
          onAxisNumber = src
        end if
      end if
    end do
if (.not. oldListOnaxisExists) call error('noAttSrc', 'There is &
&no pointing source in the old RGS source list.')

!****** other checks on old list??? Check labels are unique, and error if not!!
!****** check matchup between PROCESS/BKG_EXCLUDE and the appropriate region extensions!

  end subroutine checkOldList

  !---------------------------------------------------------------------
  subroutine readEpicData(epicSrcTab, newEpicSrcData, lastEpicSrc)
  !---------------------------------------------------------------------

    implicit none 

    type(TableT), intent(in) ::&
      epicSrcTab
    type(sourceType), dimension(:), intent(inout) ::&
      newEpicSrcData
    integer(int16), intent(out) ::&
      lastEpicSrc

    integer(int16) ::&
      maxNumNewEpicSrcs
    integer(int16) ::&
      src
    real(single), dimension(:), pointer ::&
      epicRatePtr, epicCountsPtr, epicBkgRatePtr
    real(double), dimension(:), pointer ::&
      epicRaPtr, epicDecPtr
    integer(int32), dimension(:), pointer ::&
      epicIndexPtr, epicBandPtr

    !-------------------------------------------------------------------
    ! Start:

    maxNumNewEpicSrcs = size(newEpicSrcData)
    call initSrcDataStructure(newEpicSrcData)

    epicRaPtr => real64Data(column(epicSrcTab, 'RA'))
    epicDecPtr => real64Data(column(epicSrcTab, 'DEC'))
    if (hasColumn(epicSrcTab, 'ML_ID_SRC')) then
      epicIndexPtr => int32Data(column(epicSrcTab, 'ML_ID_SRC'))
    else if (hasColumn(epicSrcTab, 'BOX_ID_SRC')) then
      epicIndexPtr => int32Data(column(epicSrcTab, 'BOX_ID_SRC'))
    else
call error('badEpicIdColumn', 'The supplied EPIC source list has neither &
&a ML_ID_SRC nor a BOX_ID_SRC column.')
    end if
    epicBandPtr => int32Data(column(epicSrcTab, 'ID_BAND'))
    epicRatePtr => real32Data(column(epicSrcTab, 'RATE'))
epicCountsPtr => real32Data(column(epicSrcTab, 'SCTS'))
epicBkgRatePtr => real32Data(column(epicSrcTab, 'BG_MAP'))

    lastEpicSrc = 0
    do src = 0, maxNumNewEpicSrcs - 1
      if (epicBandPtr(src) .eq. epicBandNum) then
        lastEpicSrc = lastEpicSrc + 1

!        newEpicSrcData(lastEpicSrc)%raDeg = real(epicRaPtr(src), single)
!        newEpicSrcData(lastEpicSrc)%decDeg = real(epicDecPtr(src), single)
newEpicSrcData(lastEpicSrc)%raDeg = epicRaPtr(src)
newEpicSrcData(lastEpicSrc)%decDeg = epicDecPtr(src)
        newEpicSrcData(lastEpicSrc)%rate = epicRatePtr(src)
newEpicSrcData(lastEpicSrc)%counts = epicCountsPtr(src)
newEpicSrcData(lastEpicSrc)%bkgRate = epicBkgRatePtr(src)
newEpicSrcData(lastEpicSrc)%bkgFlux = epicBkgRatePtr(src) / pixelAreaArcsec

        write(newEpicSrcData(lastEpicSrc)%label, '(a4, i5.5)')&
        trim(epicLabelPrefix), int(epicIndexPtr(src), int16)
      end if
    end do

!**** get newEpicTrueFilename
!newEpicTrueFilename = "don't know!"

  end subroutine readEpicData

  !---------------------------------------------------------------------
  function spatialFilter(epicSrcTab) result(numOfRows)
  !---------------------------------------------------------------------

    use various_mod

    implicit none 

    type(TableT), intent(inout) ::&
      epicSrcTab
    integer(int16) ::&
      numOfRows

    !-------------------------------------------------------------------
    ! Start:

    if (enableSpatialFilter) then
!fovFilterStr = '(DELTA_XDSP in [-3.0:3.0])'
      call initFilter(fovFilterStr)
      numOfRows = doFilter(epicSrcTab)

      if (numOfRows .le. 0) then
        call warning('noEpicSourcesInFov', 'None of the epic sources &
        &were found to lie within the field of view of the RGS camera.')
      end if

    else
      numOfRows = numberOfRows(epicSrcTab)
      if (numOfRows .le. 0) then
        call warning('noEpicSources', 'There are no epic sources in &
        &the correct energy band.')
      end if
    end if

  end function spatialFilter

  !---------------------------------------------------------------------
  subroutine getUserSrc(userSrcData)
  !---------------------------------------------------------------------

    implicit none 

    type(sourceType), intent(out) ::&
      userSrcData

    !-------------------------------------------------------------------
    ! Start:

    if (userSrcLabel .eq. '') then
call error('emptyUserLabel', 'The label for the user-specified source is &
&not permitted to consist only of blanks.')
    end if

!***** NOTE!! The following checks are designed mostly to detect when the user has allowed the ra, dec, deltadisp etc parameters to take their default values. The user should not do this. The checks depend to some degree on the values of the relevant parameter defaults, hence IF THESE ARE CHANGED THIS ALGORITHM MAY ALSO NEED TO BE CHANGED.
    if (userSrcStyle .eq. 'radec') then
      if (userSrcRaDeg .lt. 0.0 .or. userSrcRaDeg .gt. 360.0) then
call error('illegalUserRa', 'The --ra for the user-specified source is either &
&< 0  or > 360 degrees. Have you forgotten to set it?')
      end if
      if (userSrcDecDeg .lt. -90.0 .or. userSrcDecDeg .gt. 90.0) then
call error('illegalUserDec', 'The --dec for the user-specified source is &
&either < -90 or > 90 degrees. Have you forgotten to set it?')
      end if

    else ! userSrcStyle .eq. 'wrtatt'
      if (userSrcXdspArcmin .eq. 0.0 .or. userSrcDispArcmin .eq. 0.0) then
        call warning('zeroWrtAttCoord', 'Either --deltaxdsp or --deltadisp &
        &has been left at its default value of 0 arcminutes. Have you &
        &forgotten to set it?')
      end if
    end if

    call initSrcDataStructure(userSrcData)

!    userSrcLabel = no0s(userSrcLabel)
    userSrcData%label = userSrcLabel
    if (userSrcStyle .eq. 'radec') then
      userSrcData%raDeg  = userSrcRaDeg
      userSrcData%decDeg = userSrcDecDeg
      userSrcData%fixedOnSky = .true.

    else ! userSrcStyle .eq. 'wrtatt'
      userSrcData%deltaDispArcmin = userSrcDispArcmin
      userSrcData%deltaXdspArcmin = userSrcXdspArcmin
      call dispXdspToPolar(&
        userSrcData%deltaDispArcmin,&
        userSrcData%deltaXdspArcmin,&
        userSrcData%fovPhiDeg,&
        userSrcData%fovRdeg&
      )
      userSrcData%fixedOnSky = .false.
    end if
    userSrcData%process = processUserSrc
    userSrcData%bkgExclude = bkgExcludeUserSrc
    userSrcData%rate = userSrcRate

  end subroutine getUserSrc

  !---------------------------------------------------------------------
  subroutine readTable(tab, srcData)
  !---------------------------------------------------------------------

    implicit none 

    type(TableT), intent(in) ::&
      tab
    type(sourceType), dimension(:), intent(out) ::&
      srcData

    integer(int16) ::&
      numSrcs
    integer(int32) ::&
      src
    type(ColumnT) ::&
      indexCol,&
      labelCol,&
      raCol,&
      decCol,&
      rateCol,&
      dispCol,&
      xdspCol,&
      phiCol,&
      rCol,&
      confnCol,&
      procCol,&
      excluCol,&
      fixedCol,&
      epicnCol
    integer(int16), dimension(:), pointer ::&
      indexColPtr
    integer(int32), dimension(:), pointer ::&
      epicnColPtr
    real(single), dimension(:), pointer ::&
      raColPtr,&
      decColPtr,&
      rateColPtr,&
      dispColPtr,&
      xdspColPtr,&
      phiColPtr,&
      rColPtr,&
      confnColPtr
    logical(kind=bool), dimension(:), pointer ::&
      procColPtr,&
      excluColPtr,&
      fixedColPtr

    !-------------------------------------------------------------------
    ! Start:

    numSrcs = size(srcData)

    ! Open the columns:
    indexCol = column(tab, 'INDEX')
    labelCol = column(tab, 'LABEL')
    raCol    = column(tab, 'RA')
    decCol   = column(tab, 'DEC')
    rateCol  = column(tab, 'RATE')
    dispCol  = column(tab, 'DELTA_DISP')
    xdspCol  = column(tab, 'DELTA_XDSP')
    phiCol   = column(tab, 'FOV_PHI')
    rCol     = column(tab, 'FOV_R')
    confnCol = column(tab, 'CONFUSION')
    procCol  = column(tab, 'PROCESS')
    excluCol = column(tab, 'BKG_EXCLUDE')
    fixedCol = column(tab, 'FIXED_ON_SKY')
    epicnCol = column(tab, 'EPIC_FILE')

    ! Open column data values:
    indexColPtr => int16Data( indexCol)
    raColPtr    => real32Data(   raCol)
    decColPtr   => real32Data(  decCol)
    rateColPtr  => real32Data( rateCol)
    dispColPtr  => real32Data( dispCol)
    xdspColPtr  => real32Data( xdspCol)
    phiColPtr   => real32Data(  phiCol)
    rColPtr     => real32Data(    rCol)
    confnColPtr => real32Data(confnCol)
    procColPtr  => boolData(   procCol)
    excluColPtr => boolData(  excluCol)
    fixedColPtr => boolData(  fixedCol)
    epicnColPtr => int32Data( epicnCol)

    ! write column values to data array:
    srcData(1:numSrcs)%index           = indexColPtr
    srcData(1:numSrcs)%raDeg           = raColPtr
    srcData(1:numSrcs)%decDeg          = decColPtr
    srcData(1:numSrcs)%rate            = rateColPtr
    srcData(1:numSrcs)%deltaDispArcmin = dispColPtr
    srcData(1:numSrcs)%deltaXdspArcmin = xdspColPtr
    srcData(1:numSrcs)%fovPhiDeg       = phiColPtr
    srcData(1:numSrcs)%fovRdeg         = rColPtr
    srcData(1:numSrcs)%confusion       = confnColPtr
    srcData(1:numSrcs)%process         = procColPtr
    srcData(1:numSrcs)%bkgExclude      = excluColPtr
    srcData(1:numSrcs)%fixedOnSky      = fixedColPtr

    do src = 1, numSrcs
      if (isNull(epicnCol, src - 1)) then
        srcData(src)%epicFileAttrNum = myInt32Null
      else
        srcData(src)%epicFileAttrNum = epicnColPtr(src - 1)
      end if

      srcData(src)%label = no0s(trim(stringCell(labelCol, src - 1)))
    end do

  end subroutine readTable

  !---------------------------------------------------------------------
  subroutine addProposalSrc(srcData)
  !---------------------------------------------------------------------

    implicit none 

    type(sourceType), intent(inout) ::&
      srcData

    type(ProposalInfoType) :: proposalInfo ! typedef from oal

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Finding proposal info from the ODF.')

    call OAL_proposalInfo(proposalInfo)
!**** presumably the state of the OAL doesn't have to be set first? It isn't guaranteed to be set at the moment anyway... eg if attstyle=user.
    srcData%raDeg  = proposalInfo%targetRa
    srcData%decDeg = proposalInfo%targetDec
    srcData%label  = proposalLabel 

    ! All others: leave as defaults.

  end subroutine addProposalSrc

  !---------------------------------------------------------------------
  subroutine addOnaxisSrc(srcData)
  !---------------------------------------------------------------------

    implicit none 

    type(sourceType), intent(inout) ::&
      srcData

    !-------------------------------------------------------------------
    ! Start:

    srcData%deltaDispArcmin = 0.0
    srcData%deltaXdspArcmin = 0.0
    call dispXdspToPolar(&
      srcData%deltaDispArcmin,&
      srcData%deltaXdspArcmin,&
      srcData%fovPhiDeg,&
      srcData%fovRdeg&
    )
    srcData%label = onAxisLabel 
    srcData%fixedOnSky = .false.
    srcData%raDeg  = rgsAttitude%ra  / piOn180
    srcData%decDeg = rgsAttitude%dec / piOn180

    ! All others: leave as defaults.

  end subroutine addOnaxisSrc

  !---------------------------------------------------------------------
  subroutine loadAndTestForLabelClashes(trialData, firstTrialSrc&
  , firmData, lastFirmSrc)
  !---------------------------------------------------------------------

    implicit none 

    integer(int16), intent(in) ::&
      firstTrialSrc
    integer(int16), intent(inout) ::&
      lastFirmSrc
    type(sourceType), dimension(:), intent(in) ::&
      trialData
    type(sourceType), dimension(:), intent(inout) ::&
      firmData
    logical :: labelFoundBefore
    integer(int16) :: trialSrc, firmSrc

    !-------------------------------------------------------------------
    ! Start:

    do trialSrc = firstTrialSrc, size(trialData)
      labelFoundBefore = .false.
      INNER: do firmSrc = 1, lastFirmSrc
        if (trialData(trialSrc)%label .eq. firmData(firmSrc)%label) then
          labelFoundBefore = .true.
          write(errstr, '(a)') 'Duplicate label: '&
          //trim(trialData(trialSrc)%label)
          if (clobberOnLabelClash) then
            call warning('duplicateLabel')
      exit INNER

          else
call error('duplicateLabel')
          end if
        end if
      end do INNER

      if (.not. labelFoundBefore) then
        lastFirmSrc = lastFirmSrc + 1
if (lastFirmSrc .gt. size(firmData)) call error('bug2', 'Contact the code developer.')
        firmData(lastFirmSrc) = trialData(trialSrc)
      end if
    end do

!***** if user source is one that clashes, should overwrite the other source, not delete the user source??? probably not.

!***** this thing should return statuses, and let the calling routine do the error/warnings????

  end subroutine loadAndTestForLabelClashes

  !---------------------------------------------------------------------
  subroutine writeToDataSet(srcData, instrumentName, newEpicAttrNum&
  , instExpIdStr, set, tab)
  !---------------------------------------------------------------------

    implicit none 

    type(sourceType), dimension(:), intent(inout) ::&
      srcData
    character(len=*) :: instrumentName
    integer(int32), intent(in) :: newEpicAttrNum
    character(*), intent(in) :: instExpIdStr
    type(DataSetT), intent(out) :: set
    type(TableT), intent(out) :: tab

    integer(int16) ::&
      numSrcs

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Writing the source list to the outset.')

    numSrcs = size(srcData)

    select case(trim(fileMode))
    case('create')
      ! Open file for write:
      set = dataSet(newListSetName, CREATE)

      call writeHeaderKeywords(set, trim(instrumentName), instExpIdStr)

      tab = addTable(set, 'SRCLIST', int(numSrcs))

      if (withEpicSet) then
        call writeEpicSpecificKwds(tab, newEpicAttrNum)
      end if

      call writeNewTable(tab, srcData)

    case('modify')
      ! Open file for modify:
      set = dataSet(oldListSetName, MODIFY)

      if (changeAttitude) then
        call deleteRegionExtensions(set)
        srcData%process = .false.
        srcData%bkgExclude = .false.

        call setAttribute(set, 'RA_PNT',  rgsAttitude%ra   / piOn180)
        call setAttribute(set, 'DEC_PNT', rgsAttitude%dec  / piOn180)
        call setAttribute(set, 'PA_PNT',  rgsAttitude%apos / piOn180)
      end if

      tab = table(set, 'SRCLIST')

      if (withEpicSet) then
        call writeEpicSpecificKwds(tab, newEpicAttrNum)
      end if

      call deleteRows(tab, 0, numberOfRows(tab) - 1)
      call insertRows(tab, 0, int(numSrcs) - 1)
      call writeOldTable(tab, srcData)

    case default
call error('badFileMode', 'The value '//trim(fileMode)//' of parameter &
&--filemode is not recognised.')
    end select

    ! Leaves set, tab open on return.

  end subroutine writeToDataSet

  !---------------------------------------------------------------------
  subroutine reportOofovFlags(tab, srcData)
  !---------------------------------------------------------------------

    implicit none 

    type(TableT), intent(in) :: tab
    type(sourceType), dimension(:), intent(inout) ::&
      srcData

    integer(int16) ::&
      src, numSrcs
    type(ColumnT) ::&
      flagCol
    integer(int32), dimension(:), pointer ::&
      flagColPtr
character(10) :: indexStr
integer :: status

    !-------------------------------------------------------------------
    ! Start:

    numSrcs = size(srcData)

    flagCol  = column(tab, 'FLAG')
    flagColPtr  => int32Data( flagCol)
    srcData%flag = flagColPtr
    call release(flagCol)

    ! Check if any srcData%flags have been set to OUTOFFOV_MASK and warn if so.
    do src = 1, numSrcs
!      if (srcData(src)%flag .eq. 1) then
if (btest(srcData(src)%flag, OUTOFFOV_BIT)) then
call intToStr(int(src), 'left', indexStr, status)
!        write(errstr, '(a, i5, a)') 'Source INDEX=', src&
write(errstr, '(a, i5, a)') 'Source INDEX='//trim(indexStr)&
//' ('//trim(srcData(src)%label)//') is outside the field of &
!        , ' is outside the field of &
        &view of the RGS camera.'
!***** may need to alter this when and if flag column comes into being in the output source list. In that case this loop should also be outside the 'if(enable etc)' in the calling level.
        call warning('outsideFOV')
      end if
    end do

  end subroutine reportOofovFlags

  !---------------------------------------------------------------------
  subroutine getEpicAttData(epicInstrumentId, epicTimeStamp)
  !---------------------------------------------------------------------

    implicit none 

    integer(int32), intent(out) :: epicInstrumentId
    real(double),   intent(out) :: epicTimeStamp

    type(DataSetT) :: set

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(epicSrcSetName, READ)
    epicInstrumentId = instId(stringAttribute(set, 'INSTRUME'))
    epicTimeStamp = toTimeTag(stringAttribute(set, 'DATE-OBS'))

    if (.not. haveScAttitude) then
!      call warning('', '')

      scAttitude%ra   = real64Attribute(set, "RA_PNT")  * piOn180
      scAttitude%dec  = real64Attribute(set, "DEC_PNT") * piOn180
      scAttitude%apos = real64Attribute(set, "PA_PNT")  * piOn180
!***** NOTE!!! these kwds OUGHT to contain the epic BORESIGHT-CORRECTED attitude. However they seem at the moment to contain the median values from the atthkgen file.
    end if

    call release(set)

  end subroutine getEpicAttData

  !---------------------------------------------------------------------
  subroutine calcEpicPlaneConfusion(epicInstrumentId, epicTimeStamp, srcData)
  !---------------------------------------------------------------------

    use my_psf_mod

    implicit none 

    integer(int32), intent(in) ::&
      epicInstrumentId
    real(double),   intent(in) ::&
      epicTimeStamp
    type(sourceType), dimension(:), intent(inout) ::&
      srcData

    type(srcEnvelopeType) ::&
      subsetSrcData(size(srcData))
    integer(int16) ::&
      i,&
      j,&
      numSrcs,&
      numSubsetSrcs,&
status
    real(double),  dimension(:), allocatable ::&
      subsetSrcXdet,&
      subsetSrcYdet
    real(single),  dimension(:), allocatable ::&
      srcXdet,&
      srcYdet,&
      srcFovRadiusDet,&
!      subsetSrcXdet,&
!      subsetSrcYdet,&
      subsetSrcRadiusDet
    integer(int16),  dimension(:), allocatable ::&
      subsetConfSetNum

    !-------------------------------------------------------------------
    ! Start:

    numSrcs = size(srcData)

    allocate(srcXdet(numSrcs), srcYdet(numSrcs), subsetSrcXdet(numSrcs)&
    , subsetSrcYdet(numSrcs), srcFovRadiusDet(numSrcs))

!    call transformSrcPosnsToDetXY(srcData%raDeg, srcData%decDeg&
!    , epicInstrumentId, scAttitude, epicTimeStamp, srcXdet, srcYdet&
!    , srcFovRadiusDet)

call CAL_setState(instrument=epicInstrumentId)
call raDecToDetXY(srcData%raDeg, srcData%decDeg, scAttitude&
, epicTimeStamp, srcXdet, srcYdet, status=status)

if (status .eq. -1) then
call error('bug11', 'Contact the code developer.')
else if (status .eq. -2) then
call error('bug12', 'Contact the code developer.')
end if

srcFovRadiusDet = sqrt(srcXdet * srcXdet + srcYdet * srcYdet)

!print *, ''
    ! Get subset of those srcs with flag = 0
    subsetSrcData%EeV = (bandMinEeV + bandMaxEeV) / 2.0
    numSubsetSrcs = 0
    do i = 1, numSrcs
      if (srcData(i)%flag .eq. 0 .and. srcData(i)%label .ne. 'ONAXIS') then
        numSubsetSrcs = numSubsetSrcs + 1
!        subsetSrcData(numSubsetSrcs)%srcRate = srcData(i)%rate
!        subsetSrcData(numSubsetSrcs)%bkgRate = srcData(i)%bkgRate
subsetSrcData(numSubsetSrcs)%bkgFlux = srcData(i)%bkgFlux
subsetSrcData(numSubsetSrcs)%thetaArcsec&
        = srcFovRadiusDet(i) * detUnitArcsec

subsetSrcData(numSubsetSrcs)%peakFlux = srcData(i)%counts&
/ getPsfIntegral(subsetSrcData(numSubsetSrcs)%thetaArcsec&
, subsetSrcData(numSubsetSrcs)%EeV)
! in counts per square arc second

!print *, '  i, numSubsetSrcs, thetaArcsec:'&
!, i, numSubsetSrcs, subsetSrcData(numSubsetSrcs)%thetaArcsec
        subsetSrcXdet(numSubsetSrcs) = srcXdet(i)
        subsetSrcYdet(numSubsetSrcs) = srcYdet(i)
      end if
    end do
!print *, ''

    allocate(subsetSrcRadiusDet(numSubsetSrcs)&
    , subsetConfSetNum(numSubsetSrcs))

    call getPsfEllipseAxes(subsetSrcData(1:numSubsetSrcs)&
    , bkgFrac=fracBackground)
!print *, ''
!do i = 1, numSubsetSrcs
!print *, '  subsetSrcData(i)%longAxisArcsec:', subsetSrcData(i)%longAxisArcsec
!end do
!print *, ''

    ! Take long axes as src radii
    subsetSrcRadiusDet = subsetSrcData(1:numSubsetSrcs)%longAxisArcsec&
    / detUnitArcsec
!**** Simpler to start with to try:   subsetSrcRadiusDet = 5.0 / detUnitArcsec

!    call findConfusedSets(subsetSrcXdet(1:numSubsetSrcs)&
!    , subsetSrcYdet(1:numSubsetSrcs), subsetSrcRadiusDet(1:numSubsetSrcs)&
!    , subsetConfSetNum)
call findConfusedSets(subsetSrcXdet(1:numSubsetSrcs)&
, subsetSrcYdet(1:numSubsetSrcs), subsetSrcRadiusDet(1:numSubsetSrcs)&
, subsetConfSetNum, status)

if (status .eq. -1) then
call error('bug3', 'Contact the code developer.')
else if (status .eq. -2) then
call error('bug4', 'Contact the code developer.')
end if

!print *, ''
!do i = 1, numSubsetSrcs
!print *, '  i, subsetConfSetNum(i):', i, subsetConfSetNum(i)
!end do
!print *, ''

!print *, ''
    ! Write srcData%confusionSet values
    srcData%confusionSet = 0 ! default (indicates no confusion)
    j = 0
    do i = 1, numSrcs
!      if (srcData(i)%flag .eq. 0) then
if (srcData(i)%flag .eq. 0 .and. srcData(i)%label .ne. 'ONAXIS') then
        j = j + 1
        srcData(i)%confusionSet = subsetConfSetNum(j)
!print *, '  i, srcData(i)%label, srcData(i)%confusionSet:'&
!, i, srcData(i)%label, srcData(i)%confusionSet
      end if
    end do
!print *, ''

  end subroutine calcEpicPlaneConfusion

  !---------------------------------------------------------------------
  subroutine flagConfusedSources(tab, srcData)
  !---------------------------------------------------------------------
  ! The basic idea here is to look at each confused set and flag all but
  ! the brightest. However the procedure is slightly different in the
  ! case that the proposal src is a member of a confused set. The most
  ! common physical situation in which this can occur is if the source
  ! is bright enough to cause pileup in epic. In this case the centre of
  ! the source (where the proposal position is) will be burnt out and the
  ! epic source detection chain will detect multiple sources in a ring
  ! around the true position, instead of just a single source approximately
  ! aligned with the proposal. In this pileup case, hopefully all the
  ! sources (ie false epic detections as well as the proposal) will be
  ! confused. In this case the proposal position is best since rgs is
  ! not subject to significant pileup.
  !
  ! However the above case needs to be distinguished from the case of a
  ! source which is not piled up. In this case the danger to watch out
  ! for is the proposal source position being slightly incorrect (this
  ! can easily happen, since it is entered by the observer). So in the
  ! case that the proposal is confused with just one other epic source, the
  ! epic source is chosen instead for non-flagging.
  !
  ! If the source is piled-up but the proposal position incorrect, then
  ! the rgs has no way to find the true centre-source position.

    implicit none 

    type(TableT), intent(in) :: tab
    type(sourceType), dimension(:), intent(inout) ::&
      srcData

    type(ColumnT) :: flagCol
    integer(int32), pointer :: flagColPtr(:)
    integer(int16) ::&
      numSrcs, maxConfusionIndex, confSetNum, numSourcesInSet, brightestIndex(1)
!integer :: i
    !-------------------------------------------------------------------
    ! Start:

    flagCol  = column(tab, 'FLAG')
    flagColPtr  => int32Data(flagCol)
    srcData%flag = flagColPtr

    numSrcs = size(srcData)
    if (numSrcs .ne. numberOfRows(tab))&
call error('bug10', 'Should never happen - contact the code developer.')

    maxConfusionIndex = maxval(srcData%confusionSet)
    if (maxConfusionIndex .le. 0) then
      call message(AppMsg, Verbose, 'No confused epic sources detected.')
  return
    end if

!print *, ''
    do confSetNum = 1, maxConfusionIndex
      numSourcesInSet = count(srcData%confusionSet .eq. confSetNum&
      .and. srcData%flag .eq. 0) ! test for flag is redundant but
      ! let's play safe.
!print *, '  confSetNum, numSourcesInSet:', confSetNum, numSourcesInSet

if (numSourcesInSet .eq. 1) call error('bug5', 'Contact the code developer.')

    if (numSourcesInSet .le. 0) cycle ! Shouldn't happen with present
    ! structure of findConfusedSets() but let's play safe.

      if (any(srcData%label .eq. proposalLabel .and. srcData%flag .eq. 0&
      .and. srcData%confusionSet .eq. confSetNum)) then
!print *, '  proposal is a member of this set.'
        if (numSourcesInSet .eq. 2) then ! a nice clean detection. Flag the
        ! proposal and leave the epic position unflagged:

          where(srcData%label .eq. proposalLabel)&
            srcData%flag = ior(srcData%flag, CONFUSED_MASK)

        else ! probably piled-up or otherwise with messy psf and thus
        ! accompanied by a cluster of false detections. Flag all _but_
        ! the proposal:

          where(srcData%label .ne. proposalLabel&
          .and. srcData%confusionSet .eq. confSetNum)&
            srcData%flag = ior(srcData%flag, CONFUSED_MASK)
        end if

      else ! PROPOSAL is not a member of the confused set. Flag all but
      ! the brightest.
!print *, '  proposal is NOT a member of this set.'
        brightestIndex = maxloc(srcData%rate, mask=(srcData%flag .eq. 0&
        .and. srcData%confusionSet .eq. confSetNum))

        if (srcData(brightestIndex(1))%rate .le. 0.0) then
          call warning('confSetAllZeroRate', 'A set of confused sources &
          &has none with a positive value of RATE. All have been flagged.')

          where(srcData%confusionSet .eq. confSetNum)&
            srcData%flag = ior(srcData%flag, CONFUSED_MASK)

        else
          where(srcData%index .ne. srcData(brightestIndex(1))%index&
          .and. srcData%confusionSet .eq. confSetNum)&
            srcData%flag = ior(srcData%flag, CONFUSED_MASK)
        end if
      end if
!print *, ''
!do i = 1, numSrcs
!print *, '    i, srcData(i)%label, srcData(i)%confusionSet, srcData(i)%flag:'&
!, i, srcData(i)%label, srcData(i)%confusionSet, srcData(i)%flag
!end do
!print *, ''
    end do
!print *, ''

    flagColPtr = srcData%flag 
    call release(flagCol)

  end subroutine flagConfusedSources

  !---------------------------------------------------------------------
  function calcPrimeIndex(oldListTab, srcData) result(indexOfPrime)
  !---------------------------------------------------------------------

    implicit none 

    type(sourceType), dimension(:), intent(in) ::&
      srcData
    type(TableT), intent(in) :: oldListTab

    integer(int16) ::&
      numSrcs
!    type(DataSetT) :: oldListSet
    integer(int16) ::&
      indexOfPrime, oldIndexOfPrime, proposalIndex, pointingIndex, src, dummy(1), indexOfBrightest
    real(single) :: valueOfBrightest
    logical :: notFound

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Calculating the INDEX value of the &
    &prime source.')

    ! NOTE!!! The algorithm in this subroutine relies for some extent on the parameter defaults, hence IF THESE ARE CHANGED THIS ALGORITHM MAY ALSO NEED TO BE CHANGED.

    numSrcs = size(srcData)

!.......................................................................
! Get some index values which might be useful:

    ! Get the previous prime source number if there is one:
    if (trim(fileMode) .eq. 'modify') then
      oldindexOfPrime = int16Attribute(oldListTab, 'PRIMESRC')
    end if

    ! Get the index of the proposal source (better not assume it is 1):
    notFound = .true. ! default
    do src = 1, numSrcs
      if (srcData(src)%label .eq. proposalLabel) then
        notFound = .false.
        proposalIndex = srcData(src)%index
    exit
      end if
    end do

if (notFound) call error('bug6', 'No proposal source found. Contact the code developer.')

    ! Get the index of the pointing source (better not assume it is 2):
    notFound = .true. ! default
    do src = 1, numSrcs
      if (srcData(src)%label .eq. onAxisLabel) then
        notFound = .false.
        pointingIndex = srcData(src)%index
    exit
      end if
    end do

if (notFound) call error('bug7', 'No pointing source found. Contact the code developer.')

!.......................................................................
! Start of main prime source choice. Mostly the next section is taken up by
! the following: (fileMode .eq. 'create' .or. changePrimeSrc) .and. addUserSrc
! .and. setUserSrcToPrime.

indexOfPrime = -1 ! an impossible default, to pick up bugs in this routine. Eventually this line (and the tests at the end of the sub) could probably be dispensed with.

    if (trim(fileMode) .eq. 'modify' .and. .not. changePrimeSrc) then
      indexOfPrime = oldindexOfPrime

    else ! either fileMode .eq. 'create' .or. changePrimeSrc
      if (addUserSrc .and. setUserSrcToPrime) then
        ! Find user source:
        notFound = .true. ! default
        do src = 1, numSrcs
          if (srcData(src)%label .eq. userSrcLabel) then
            notFound = .false.
            indexOfPrime = srcData(src)%index
        exit
          end if
        end do
if (notFound) call error('bug8', 'Contact the code developer.') ! in current implementation user _label_ should always be present, even if there was a label clash with another src and the _actual_ user source has been discarded. In this case indexOfPrime is _not_ going to point to the user source but the existing source that had the same label value. Hence in no case should this error occur.

!.......................................................................
! Remainder of the choice tree deals with non-user prime source cases.

      else ! either no user source supplied, or it wasn't requested to be
      ! the prime source. Look now at the dedicated prime source parameters.
        ! Calculate indexOfPrime:
        select case(primeSrcStyle)
        case('label')
!**** check for all-blank primeSrcLabel? Remember that all-blank labels are not allowed. Same with epic label prefix???
          notFound = .true.
          do src = 1, numSrcs
            if (srcData(src)%label .eq. primeSrcLabel) then
              notFound = .false.
              indexOfPrime = srcData(src)%index
          exit
            end if
          end do

          if (notFound) then
            indexOfPrime = proposalIndex
            call warning('primeLabelNotFound', 'Label '//trim(primeSrcLabel)&
            //' not found; prime source set to proposal source.')
          end if

        case('index')
          notFound = .true.
          do src = 1, numSrcs
            if (srcData(src)%index .eq. paramIndexOfPrime) then
              notFound = .false.
              indexOfPrime = srcData(src)%index
          exit
            end if
          end do

          if (notFound) then
            indexOfPrime = proposalIndex
            write(errstr, '(a, i5, a)') 'Index ', paramIndexOfPrime&
            , ' not found; prime source set to proposal source.'
            call warning('primeIndexNotFound')
          end if

        case('expression')
call error('notSupported', 'Sorry, this mode of entering the prime source &
&is not yet supported.')

        case('brightest')
          dummy = maxloc(srcData%rate)
!***** mask it by flag > 0??? Should also consider impact of flag column on other prime source cases - eg generate some warnings if the prime source is flagged?
          indexOfPrime = dummy(1)
          if (srcData(indexOfPrime)%rate .le. 0.0) then
            indexOfPrime = proposalIndex
            call warning('noBrightPrime', 'The brightest source has a rate &
            &less than or equal to zero; prime source set to proposal source.')
          end if

        case('auto') ! to be used by the pcms ('auto' should probably also
        ! be the default value of this parameter). In this case the code
        ! makes every possible attempt to ensure that the most sensible
        ! possible source is chosen to be the prime src.

          if (srcData(proposalIndex)%flag .eq. 0) then ! ie, if proposal src
          ! is neither confused nor outside the fov:

            indexOfPrime = proposalIndex

          else ! choose the brightest epic source:
            indexOfBrightest = -1 ! nonsense default.
            valueOfBrightest = 0.0
            do src = 1, numSrcs
              if (  srcData(src)%epicFileAttrNum .ne. myInt32null&
              .and. srcData(src)%flag            .eq. 0&
              .and. srcData(src)%rate            .gt. valueOfBrightest) then
                indexOfBrightest = src
                valueOfBrightest = srcData(src)%rate
              end if
!*** can do this via masked maxloc?
            end do
            if (valueOfBrightest .gt. 0.0) then
              indexOfPrime = indexOfBrightest

            else ! if all else fails, choose the pointing direction!
              indexOfPrime = pointingIndex
            end if
          end if

        case default
call error('badPrimeStyle', 'The value of parameter --primestyle is &
&not recognized.')
        end select ! primeSrcStyle
      end if ! addUserSrc .and. setUserSrcToPrime
    end if ! fileMode .eq. 'modify' .and. .not. changePrimeSrc

!.......................................................................
! Finally, do some checks on defaults etc.

    if (addUserSrc .and. setUserSrcToPrime .and. .not. changePrimeSrc) then
      call warning('noChangePrime', 'You have requested that the user &
      &source be the prime source, but have not enabled a change in &
      &prime source via the parameter --changeprime. Prime source not &
      &changed.')
    end if

    if (fileMode .eq. 'modify' .and. changePrimeSrc) then
      if (primeSrcStyle .eq. 'label' .and. primeSrcLabel .eq. 'PROPOSAL') then
        write(errstr, '(a, i5, a)') 'You have requested that the &
        &prime source be changed via the parameter --changeprime, but you &
        &may have forgotten to specify the new prime source, because &
        &it is still set to its default.'

        call warning('primeAtLabelDefault')
      end if

      if (oldindexOfPrime .eq. indexOfPrime) then
        write(errstr, '(a, i5, a)') 'You have requested that the &
        &prime source be changed via the parameter --changeprime, but you &
        &may have forgotten to specify the new prime source index, because &
        &the new prime source index of ', indexOfPrime, ' is the same as &
        &the old one.'

        call warning('primeAtIndexDefault')
      end if

    else if (fileMode .eq. 'create'&
    .and. .not. (addUserSrc .and. setUserSrcToPrime)) then
      if (primeSrcStyle .eq. 'label' .and. primeSrcLabel .eq. 'PROPOSAL') then
        write(errstr, '(a, i5, a)') 'Just a reminder that you may want &
        &to specify the prime source - it is still set to its default.'

        call warning('primeAtDefault')
      end if
    end if

    if (srcData(indexOfPrime)%flag .ne. 0) then
      call warning('primeOutsideFOV', 'The prime source is outside the &
      &field of view of the RGS camera. This will have a deleterious &
      &effect on the attitude drift correction of the RGS events.')
!***** may need to alter this when and if flag column comes into being in the output source list.
    end if

if (indexOfPrime .le. 0) call error('bug9', 'Contact the code developer.') ! to pick up bugs. Eventually this test (and the default-setting line) could probably be dispensed with.

  end function calcPrimeIndex

  !---------------------------------------------------------------------
   subroutine writeHeaderKeywords(set, instrumentName, instExpIdStr)
  !---------------------------------------------------------------------

    implicit none 

    type(DataSetT), intent(in) :: set
    character(len=*), intent(in) :: instrumentName
    character(*), intent(in) :: instExpIdStr

    integer(int32) :: instrumentId

    !-------------------------------------------------------------------
    ! Start:

    call writeGenericAttributes(set, instrumentName, rgsAttitude)
    ! 'TELESCOP'
    ! 'INSTRUME'
    ! 'FILTER'
    ! 'ORIGIN'
    ! 'RA_PNT'
    ! 'DEC_PNT'
    ! 'PA_PNT'
    ! 'RADECSYS'
    ! 'EQUINOX'
!    call setAttribute(set, 'EXPOSURE',)
!    call setAttribute(set, 'DATE-OBS',)
!    call setAttribute(set, 'DATE-END',)
!    call setAttribute(set, 'OCB',) ! special RGS header kwd: on-chip binning
!***** ...but the latter is ccd-specific. Tell Julian it should be dropped from the icd-specified list for rgs source lists.

    if (writeObsAttributes) call writeObservationAttributes(set)
    ! 'OBS_MODE'
    ! 'OBS_ID'
    ! 'OBJECT'
    ! 'OBSERVER'

    if (writeExpAttributes) then
      instrumentId = instId(instrumentName)
      call writeExposureAttributes(set, instrumentId, instExpIdStr)
    ! 'DATAMODE'
    ! 'EXP_ID'
    ! 'EXPIDSTR'
    end if

  end subroutine writeHeaderKeywords

  !---------------------------------------------------------------------
  subroutine writeEpicSpecificKwds(outTab, newEpicAttrNum)
  !---------------------------------------------------------------------

    implicit none 

    type(TableT), intent(in)   :: outTab
    integer(int32), intent(in) :: newEpicAttrNum

    type(DataSetT) :: epicSrcSet
    character(len=8) ::&
      epicExpIdAttrName,&
      epicContentAttrName,&
      epicBandAttrName,&
      epicFilterAttrName,&
      epicExpId,&
      epicFilter
!**** why call it expId when it gets he instrument name?
!**** is 8 char enough for the filter?
    character(len=80) ::&
      epicContent

    !-------------------------------------------------------------------
    ! Start:

    epicSrcSet = dataSet(epicSrcSetName, READ)
      epicExpId   = stringAttribute(epicSrcSet, 'INSTRUME')
      epicContent = stringAttribute(epicSrcSet, 'CONTENT')
      epicFilter  = stringAttribute(epicSrcSet, 'FILTER')
    call release(epicSrcSet)

    write(epicExpIdAttrName,   '(a6, i2)') 'E_EXPR', newEpicAttrNum
    write(epicContentAttrName, '(a6, i2)') 'E_CONT', newEpicAttrNum
    write(epicBandAttrName,    '(a6, i2)') 'E_BAND', newEpicAttrNum
    write(epicFilterAttrName,  '(a6, i2)') 'E_FILT', newEpicAttrNum
    call setAttribute(outTab, epicExpIdAttrName,   epicExpId)
    call setAttribute(outTab, epicContentAttrName, epicContent)
    call setAttribute(outTab, epicBandAttrName,    epicBandNum)
    call setAttribute(outTab, epicFilterAttrName,  epicFilter)

  end subroutine writeEpicSpecificKwds

  !---------------------------------------------------------------------
  subroutine writeOldTable(tab, srcData)
  !---------------------------------------------------------------------

    implicit none 

    type(TableT), intent(in) ::&
      tab
    type(sourceType), dimension(:), intent(in) ::&
      srcData

    integer(int16) ::&
      numSrcs, src
    type(ColumnT) ::&
      indexCol,&
      labelCol,&
      raCol,&
      decCol,&
      rateCol,&
      dispCol,&
      xdspCol,&
      phiCol,&
      rCol,&
      confnCol,&
      procCol,&
      excluCol,&
      fixedCol,&
      epicnCol,&
      flagCol
    integer(int16), dimension(:), pointer ::&
      indexColPtr
    integer(int32), dimension(:), pointer ::&
      epicnColPtr,&
      flagColPtr
    real(single), dimension(:), pointer ::&
      raColPtr,&
      decColPtr,&
      rateColPtr,&
      dispColPtr,&
      xdspColPtr,&
      phiColPtr,&
      rColPtr,&
      confnColPtr
    logical(bool), dimension(:), pointer ::&
      procColPtr,&
      excluColPtr,&
      fixedColPtr

    !-------------------------------------------------------------------
    ! Start:

    numSrcs = size(srcData)

    ! Open the columns:
    indexCol = column(tab, 'INDEX')
    labelCol = column(tab, 'LABEL')
    raCol    = column(tab, 'RA')
    decCol   = column(tab, 'DEC')
    rateCol  = column(tab, 'RATE')
    dispCol  = column(tab, 'DELTA_DISP')
    xdspCol  = column(tab, 'DELTA_XDSP')
    phiCol   = column(tab, 'FOV_PHI')
    rCol     = column(tab, 'FOV_R')
    confnCol = column(tab, 'CONFUSION')
    procCol  = column(tab, 'PROCESS')
    excluCol = column(tab, 'BKG_EXCLUDE')
    fixedCol = column(tab, 'FIXED_ON_SKY')
    epicnCol = column(tab, 'EPIC_FILE')
flagCol  = addColumn(tab, 'FLAG', INTEGER32, comment = 'Quality flag.')
!***** because there isn't one yet in the disk file format.

    ! Open column data values:
    indexColPtr => int16Data( indexCol)
    raColPtr    => real32Data(   raCol)
    decColPtr   => real32Data(  decCol)
    rateColPtr  => real32Data( rateCol)
    dispColPtr  => real32Data( dispCol)
    xdspColPtr  => real32Data( xdspCol)
    phiColPtr   => real32Data(  phiCol)
    rColPtr     => real32Data(    rCol)
    confnColPtr => real32Data(confnCol)
    procColPtr  => boolData(   procCol)
    excluColPtr => boolData(  excluCol)
    fixedColPtr => boolData(  fixedCol)
    epicnColPtr => int32Data( epicnCol)
    flagColPtr  => int32Data( flagCol)

    call setNullValue(epicnCol, myInt32null)
    if (.not. nullDefined(nullable(epicnCol))) then
call error('unableToSetNull', 'Could not set an int32 null value for &
&column EPIC_FILE.')
    end if

    ! write column values:
    do src = 1, numSrcs
      indexColPtr(src - 1) = srcData(src)%index
      raColPtr(src - 1)    = srcData(src)%raDeg
      decColPtr(src - 1)   = srcData(src)%decDeg
      rateColPtr(src - 1)  = srcData(src)%rate
      dispColPtr(src - 1)  = srcData(src)%deltaDispArcmin
      xdspColPtr(src - 1)  = srcData(src)%deltaXdspArcmin
      phiColPtr(src - 1)   = srcData(src)%fovPhiDeg
      rColPtr(src - 1)     = srcData(src)%fovRdeg
      confnColPtr(src - 1) = srcData(src)%confusion
      procColPtr(src - 1)  = srcData(src)%process
      excluColPtr(src - 1) = srcData(src)%bkgExclude
      fixedColPtr(src - 1) = srcData(src)%fixedOnSky
      flagColPtr(src - 1)  = srcData(src)%flag

      if (srcData(src)%epicFileAttrNum .eq. myInt32null) then
        call setToNull(epicnCol, src - 1)
      else
        epicnColPtr(src - 1) = srcData(src)%epicFileAttrNum
      end if

      call setStringCell(labelCol , src - 1, srcData(src)%label)
! DON'T trim the string argument to setStringCell. If you do, the characters
! in the file that should be trailing blanks will not be overwritten from
! whatever they were before in the memory map of the file, and will thus
! contain nonsense characters.
    end do

    call release(indexCol)
    call release(labelCol)
    call release(raCol)
    call release(decCol)
    call release(rateCol)
    call release(dispCol)
    call release(xdspCol)
    call release(phiCol)
    call release(rCol)
    call release(confnCol)
    call release(procCol)
    call release(excluCol)
    call release(fixedCol)
    call release(epicnCol)
    call release(flagCol)

  end subroutine writeOldTable

  !---------------------------------------------------------------------
  subroutine deleteRegionExtensions(set)
  !---------------------------------------------------------------------

    implicit none 

    type(DataSetT), intent(in) :: set
    integer :: blockNumber

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Deleting all region extensions.')

    do blockNumber = 1, numberOfBlocks(set) - 1 ! start from 1 not 0 because
    ! we don't want to delete the source list!
      call deleteBlock(set, blockNumber)
    end do

  end subroutine deleteRegionExtensions

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  !---------------------------------------------------------------------
  subroutine test
  !---------------------------------------------------------------------

    implicit none 

    real(double) :: myOffAxisAngles(2)
    type(SpacecraftAttitudeType) :: attitude, attBkp
    ! type defined in caloalutils/src/caloaldefs.f90
    type(SkyPositionType) :: skyPos, newSkyPos, oldSkyPos

    real(double) :: oldOffAxisAngles(2)

    !-------------------------------------------------------------------
    ! Start:

    call CAL_setState(instrument=RGS2)

attitude%ra   =  2.1369779898568871
attitude%dec  = -0.8262013799533605
attitude%apos =  5.0222221375344533

attBkp = attitude

skyPos%ra  =  2.1369928906537610
skyPos%dec = -0.8262388705572767
!skyPos%ra  =  attitude%ra
!skyPos%dec = attitude%dec

    myOffAxisAngles = raDecToOffAxis(attitude, skyPos) 
print *, ' my offAxisAngles:', myOffAxisAngles


oldOffAxisAngles = offAxisAngles(SpacecraftAttitude=attitude, SkyPos=skyPos)
print *, 'old offAxisAngles:', oldOffAxisAngles


attBkp = attitude
    newSkyPos = offAxisToRaDec(attBkp, myOffAxisAngles)

print *,    skyPos%ra,    skyPos%dec
print *, newSkyPos%ra, newSkyPos%dec
print *, ' '

attBkp = attitude
    oldSkyPos = offAxisToRaDec(attBkp, oldOffAxisAngles)
print *,    skyPos%ra,    skyPos%dec
print *, oldSkyPos%ra, oldSkyPos%dec

  end subroutine test

  !---------------------------------------------------------------------
  subroutine dumpRegion(x, y, fileName)
  !---------------------------------------------------------------------

    implicit none 

    real(single), intent(in) :: x(:), y(:)
    character(len=*), intent(in) ::&
      fileName

    integer :: n, i, ioDevNum
    character(len=255) ::&
      polygonStr

    !-------------------------------------------------------------------
    ! Start:

    n = size(x)

    write(polygonStr, '(a)') 'fk5;polygon('
    do i = 1, n - 1
      write(polygonStr, '(a, es13.6, a, es13.6, a)') trim(polygonStr)&
      , x(i), ',', y(i), ','
    end do
    write(polygonStr, '(a, es13.6, a, es13.6, a)') trim(polygonStr)&
    , x(n), ',', y(n), ')'

!***    call get_free_io_unit(ioDevNum)
ioDevNum = 1
    open(unit = ioDevNum, file = trim(fileName), status = 'replace')
      write(ioDevNum, '(a)') '# Region file format: DS9 version 3.0'
      write(ioDevNum, '(a)') "global color=green font='helvetica 10 &
      &normal' edit=1 move=1 delete=1 include=1 fixed=0"
      write(ioDevNum, '(a)') trim(polygonStr)
    close(ioDevNum)

  end subroutine dumpRegion

  !---------------------------------------------------------------------
  function calcAlpha(beta, y) result(alpha)
  !---------------------------------------------------------------------
  ! Uses the dispersion relation to calculate alpha, the angle between
  ! the incident ray and the grating, from beta, the angle between the
  ! dispersed ray and the grating, and y = order * wavelength
  ! / ruling_separation. Beta and alpha are assumed small and therefore
  ! a relation is used that avoids imprecision due to taking cosines of
  ! small angles.

    implicit none 

    real(single), intent(in) :: beta
    real(double), intent(in) :: y
    real(single) :: alpha

    !-------------------------------------------------------------------
    ! Start:

!    alpha = acos(cos(beta) - y) ! imprecise for small beta/alpha.
    alpha = asin(sqrt(sin(beta) * sin(beta) - y * y + 2.0 * y * cos(beta)))

  end function calcAlpha

  !---------------------------------------------------------------------
  subroutine dumpSrcData(srcData, numSrcs)
  !---------------------------------------------------------------------

    implicit none 

    integer(int16), intent(in) ::&
      numSrcs
    type(sourceType), dimension(numSrcs), intent(in) ::&
      srcData
    integer :: src

    !-------------------------------------------------------------------
    ! Start:

    do src = 1, numSrcs
      print *, ''
      print *, ''
      print *, 'Source ', src, ':'
      print *, ''
      print *, '       index:', srcData(src)%index
      print *, '       label:', srcData(src)%label
      print *, '          ra:', srcData(src)%raDeg
      print *, '         dec:', srcData(src)%decDeg
      print *, '        rate:', srcData(src)%rate
      print *, '  delta_disp:', srcData(src)%deltaDispArcmin
      print *, '  delta_xdsp:', srcData(src)%deltaXdspArcmin
      print *, '     fov_phi:', srcData(src)%fovPhiDeg
      print *, '       fov_r:', srcData(src)%fovRdeg
      print *, '   confusion:', srcData(src)%confusion
      print *, '     process:', srcData(src)%process
      print *, ' bkg_exclude:', srcData(src)%bkgExclude
      print *, 'fixed_on_sky:', srcData(src)%fixedOnSky
      print *, '    epic_num:', srcData(src)%epicFileAttrNum
    end do

  end subroutine dumpSrcData

  !---------------------------------------------------------------------
  subroutine transformSrcPosnsToDetXY(srcRaDeg, srcDecDeg, instrumentId&
  , myAttitude, myTimeStamp, srcXdet, srcYdet, srcFovRadiusDet)
  !---------------------------------------------------------------------
!**** Taken from region. Should be put into ssclib?
! Here the src radius calculation is taken out of the sub (and should be done so in region also?).
! Changed type of theta etc from single to double.

    implicit none

    real(single), dimension(:),   intent(in) ::&
      srcRaDeg,&
      srcDecDeg
    integer(int32),               intent(in) ::&
      instrumentId
    type(SpacecraftAttitudeType), intent(in) ::&
      myAttitude ! typedef from caloalutils
    real(double),                 intent(in) ::&
      myTimeStamp
    real(single),  dimension(:), intent(out) ::&
      srcXdet, srcYdet, srcFovRadiusDet
!, srcRadiusDet

    type(BoresightT) ::&
      boresight ! typedef from cal
    type(EquatorialDirectionType), allocatable ::&
      srcDirs(:) ! typedef from caloalutils
    real(double), dimension(:), pointer ::&
      theta, phi, cc2x, cc2y, cc2z
    real(double) ::&
      mmToDetUnits, focalLengthMm!!!, meanEeV
    integer(int16) ::&
      numSrcs, i
!    type(PsfT) ::&
!      psf

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, 'Coordinate transform.')

    call CAL_setState(instrument=instrumentId, randomize=.false.&
    , ccdModeId=PRIME_FULL_WINDOW, accuracyLevel=ACCURACY_LOW)

    boresight = CAL_getBoresight(myTimeStamp)
    call CAL_getMiscellaneousData('FOCAL_LENGTH', focalLengthMm)
    mmToDetUnits = RadToArcsec / focalLengthMm / detUnitArcsec

    numSrcs = size(srcRaDeg)
    allocate(srcDirs(numSrcs))
    do i = 1, numSrcs
      srcDirs(i)%ra = srcRaDeg(i) * piOn180
      do while(srcDirs(i)%ra .ge. twoPi)
        srcDirs(i)%ra = srcDirs(i)%ra - TwoPi
      end do
      do while(srcDirs(i)%ra .lt. 0.0)
        srcDirs(i)%ra = srcDirs(i)%ra + TwoPi
      end do
      srcDirs(i)%dec = min(max(srcDecDeg(i) * PiOn180, -PiOverTwo), PiOverTwo)
    end do

    call CAL_withSacCoordOrientation(boresight, myAttitude)
    call CAL_projectOntoTelCoord(boresight, srcDirs, theta, phi)
    call CAL_telCoordToCamCoord2(theta, phi, cc2x, cc2y, cc2z)

    srcXdet = cc2X * mmToDetUnits
    srcYdet = cc2Y * mmToDetUnits
    srcFovRadiusDet = sqrt(srcXdet * srcXdet + srcYdet * srcYdet)

!    meanEeV = (bandLoEeV(bandNum) + bandHiEeV(bandNum)) / 2.0
!    do i = 0, numSrcs - 1
!      psf = CAL_getPSF(meanEeV, dble(theta(i)), dble(phi(i)))
!      srcRadiusDet(i + 1) = CAL_psfEncircledEnergy(psf, fracEncircEnergy)&
!      * mmToDetUnits
!if (srcRadiusDet(i + 1) .lt. 0.0) call error('negativeRadius', '')
!    end do

    call CAL_releaseMemory(theta)
    call CAL_releaseMemory(phi)
    call CAL_releaseMemory(cc2x, cc2y, cc2z)

    deallocate(srcDirs)

  end subroutine transformSrcPosnsToDetXY

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

  end subroutine template
end module rgssources_mod
